package com.yanzuoguang.dao.impl;

import com.yanzuoguang.dao.DaoConst;
import com.yanzuoguang.dao.TableAnnotation;
import com.yanzuoguang.util.helper.StringHelper;
import com.yanzuoguang.util.base.MethodField;
import com.yanzuoguang.util.base.ObjectHelper;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 表结构的基本信息
 *
 * @author 颜佐光
 */
public class TableStruct {

    /**
     * 表名称
     */
    private String name;

    /**
     * 缓存的字段
     */
    private Map<Integer, List<TableFieldVo>> mapFields = new HashMap<>();

    /**
     * 获取某个类型的所有字段
     *
     * @param type 某个类型
     * @return
     */
    private List<TableFieldVo> getFieldActionList(int type) {
        if (!mapFields.containsKey(type)) {
            mapFields.put(type, new ArrayList<>());
        }
        return mapFields.get(type);
    }

    /**
     * 获取某个类型的字段
     *
     * @param type 类型
     * @return
     */
    private TableFieldVo getFieldAction(int type) {
        List<TableFieldVo> list = getFieldActionList(type);
        return list.size() > 0 ? list.get(0) : null;
    }

    /**
     * 获取所有普通子弹
     *
     * @return 所有的普通股字段
     */
    public List<TableFieldVo> getFields() {
        return getFieldActionList(DaoConst.FIELD_COMMON);
    }

    /**
     * 获取表名
     *
     * @return
     */
    public String getName() {
        return name;
    }

    /**
     * 设置表名
     *
     * @param name
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * 主键
     *
     * @return
     */
    public TableFieldVo getKey() {
        return getFieldAction(DaoConst.FIELD_PRIMARY);
    }

    /**
     * 主键名称
     *
     * @return
     */
    public String getKeyName() {
        return this.getKey().inputName;
    }

    /**
     * 主键数据类型
     *
     * @return
     */
    public Class<?> getKeyType() {
        return this.getKey().type;
    }

    /**
     * ＭＤ５缓存字段
     *
     * @return
     */
    public TableFieldVo getMd5Key() {
        return getFieldAction(DaoConst.FIELD_MD5);
    }

    /**
     * ＭＤ５缓存键值
     *
     * @return
     */
    public String getMD5KeyName() {
        TableFieldVo md5Key = this.getMd5Key();
        return md5Key != null ? md5Key.inputName : null;
    }

    /**
     * 判断是否包含版本号字段
     *
     * @return
     */
    public TableFieldVo getVersion() {
        return getFieldAction(DaoConst.FIELD_VERSION);
    }

    /**
     * 判断是否包含remove字段
     *
     * @return
     */
    public TableFieldVo getRemove() {
        return getFieldAction(DaoConst.FIELD_REMOVE);
    }


    /**
     * 获取删除时更新子弹
     *
     * @return 删除时更新字段
     */
    public List<TableFieldVo> getRemoveUpdate() {
        return getFieldActionList(DaoConst.FIELD_REMOVE_UPDATE);
    }

    /**
     * 构造函数
     */
    public TableStruct() {
        this.name = "";
    }

    /**
     * 通过实体的字段来创建表结构信息
     *
     * @param name 表名称
     * @param cls  关联的实体，主键放在第一位，其他字段放到后面；需要注意的是必需和表结构对应起来，会有隐性BUG，比如说在实体中增加了字段，会导致增加修改失败
     */
    public TableStruct(String name, Class<?> cls) {
        this.name = name;

        // 获取实体中的所有字段信息，包含get、set、field
        HashMap<String, MethodField> fields = ObjectHelper.getTypeField(cls);

        // 遍历字段
        for (Map.Entry<String, MethodField> entry : fields.entrySet()) {
            // 字段信息获取
            MethodField field = entry.getValue();
            if (field.getField() == null && field.getGetMethod() == null) {
                continue;
            }
            addMethodField(field);
        }
    }

    /**
     * 添加字段
     *
     * @param field 添加字段
     */
    private void addMethodField(MethodField field) {

        int fieldAction = DaoConst.FIELD_COMMON;

        String fieldName = field.getName();
        String fieldInputName = field.getName();
        Class<?> fieldType = String.class;

        TableAnnotation annotation = null;
        if (field.getField() != null) {
            annotation = field.getField().getAnnotation(TableAnnotation.class);
            fieldType = field.getField().getType();
        } else if (field.getGetMethod() != null) {
            annotation = field.getGetMethod().getAnnotation(TableAnnotation.class);
            fieldType = field.getGetMethod().getReturnType();
        }

        if (annotation != null) {
            fieldName = annotation.value();
            fieldAction = annotation.type();
        }

        TableFieldVo vo = new TableFieldVo(fieldName, fieldInputName, fieldType);
        int stringAction = getStringAction(vo);

        // 判断是否属于主键
        List<TableFieldVo> commonActionList = this.getFieldActionList(DaoConst.FIELD_COMMON);
        if (fieldAction == DaoConst.FIELD_PRIMARY || stringAction == DaoConst.FIELD_PRIMARY) {
            // 移除历史主键，并且将历史主键添加到普通列
            List<TableFieldVo> primaryActionList = this.getFieldActionList(DaoConst.FIELD_PRIMARY);
            commonActionList.addAll(primaryActionList);
            primaryActionList.clear();
            // 将现有列添加到主键
            primaryActionList.add(vo);
        } else {
            // 将所有非主键列添加到普通列
            commonActionList.add(vo);
            if (fieldAction == DaoConst.FIELD_REMOVE_UPDATE || stringAction == DaoConst.FIELD_REMOVE_UPDATE) {
                List<TableFieldVo> actionList = this.getFieldActionList(DaoConst.FIELD_REMOVE_UPDATE);
                actionList.add(vo);
            }
            // 处理其他特殊列
            else if (fieldAction != DaoConst.FIELD_COMMON) {
                // 假如特殊列已经存在，则将已经存在的特殊列删除，并且添加新的特殊列
                List<TableFieldVo> actionList = this.getFieldActionList(fieldAction);
                actionList.clear();
                actionList.add(vo);
            } else if (stringAction != DaoConst.FIELD_COMMON) {
                // 假如是默认的，并且特殊列已经存在，则不进行任何处理
                List<TableFieldVo> actionList = this.getFieldActionList(stringAction);
                if (actionList.isEmpty()) {
                    actionList.add(vo);
                }
            }
        }
    }

    /**
     * 获取字符串动作类型
     *
     * @param vo 输入子弹
     * @return
     */
    private int getStringAction(TableFieldVo vo) {
        if (getKey() == null) {
            return DaoConst.FIELD_PRIMARY;
        } else if (DaoConst.REMOVE_FLAG.equals(vo.inputLName)) {
            return DaoConst.FIELD_REMOVE;
        } else if (DaoConst.VERSION_FLAG.equals(vo.inputLName)) {
            return DaoConst.FIELD_VERSION;
        }
        // 判断是否属于统计字段
        else if (DaoConst.MD5_KEY_FLAG.equals(vo.inputLName)) {
            return DaoConst.FIELD_MD5;
        } else if (vo.inputLName.startsWith(DaoConst.UPDATE_FLAG)) {
            return DaoConst.FIELD_REMOVE_UPDATE;
        } else {
            return DaoConst.FIELD_COMMON;
        }

    }

    /**
     * 获取某一列
     *
     * @param name 列名
     * @return
     */
    public TableFieldVo getField(String name) {
        if (StringHelper.isEmpty(name)) {
            return null;
        }
        for (TableFieldVo fieldVo : this.getFields()) {
            if (fieldVo.lName.equals(name.toLowerCase()) || fieldVo.inputLName.equals(name.toLowerCase())) {
                return fieldVo;
            }
        }
        return null;
    }

    /**
     * 通过表结构自动生成SQL语句
     *
     * @return
     */
    public void init(TableSqlCache table) {
        table.setTable(this);
        if (!StringHelper.isEmpty(this.name)) {
            table.add(releaseSqlCreate(), releaseSqlUpdate(), releaseSqlRemove(), releaseSqlLoad());
        }
    }

    /**
     * 生成创建的SQL语句
     *
     * @return
     */
    private SqlData releaseSqlCreate() {
        // 生成添加的SQL语句
        String text = DaoConst.INSERT_MODEL.replace(DaoConst.TABLE_CODE, this.name);
        SqlData sql = new SqlData(DaoConst.CREATE, text);
        sql.setSqlType(DaoConst.SQL_TYPE_CREATE);
        String flag = "";
        if (this.getKeyType() == String.class) {
            sql.addParaConst(this.getKey().inputName, DaoConst.FIELD_CODE, this.getKey().name, DaoConst.VALUES_CODE, "?");
            flag = ",";
        }
        for (TableFieldVo field : this.getFields()) {
            sql.addParaConst(field.inputName, DaoConst.FIELD_CODE, flag + field.name, DaoConst.VALUES_CODE, flag + "?");
            flag = ",";
        }
        return sql;
    }

    /**
     * 生成修改的SQL语句
     *
     * @return
     */
    private SqlData releaseSqlUpdate() {
        // 生成添加的SQL语句
        String text = DaoConst.UPDATE_MODEL.replace(DaoConst.TABLE_CODE, this.name);
        SqlData sql = new SqlData(DaoConst.UPDATE, text);
        sql.setSqlType(DaoConst.SQL_TYPE_UPDATE);
        TableFieldVo removeField = this.getRemove();
        TableFieldVo versionField = this.getVersion();
        sql.addParaConst(this.getKey().inputName,
                DaoConst.FIELD_CODE, "a." + this.getKey().name + "=a." + this.getKey().name,
                DaoConst.WHERE_CODE, " AND a." + this.getKey().name + "=?");
        for (TableFieldVo field : this.getFields()) {
            if (field == removeField || field == versionField) {
                continue;
            }
            sql.addParaConst(field.inputName, DaoConst.FIELD_CODE, ",a." + field.name + "=?");
        }
        if (removeField != null) {
            sql.addParaConst(removeField.inputName, DaoConst.WHERE_CODE, " AND a." + removeField.name + "=0");
        }
        if (versionField != null) {
            sql.addParaConst(versionField.inputName,
                    DaoConst.FIELD_CODE, ",a." + versionField.name + "=1+a." + versionField.name,
                    DaoConst.WHERE_CODE, " AND a." + versionField.name + "=?");
        }
        return sql;
    }

    /**
     * 生成删除的SQL语句
     *
     * @return
     */
    private SqlData releaseSqlRemove() {
        TableFieldVo removeField = this.getRemove();
        TableFieldVo versionField = this.getVersion();
        if (removeField != null) {
            // 生成添加的SQL语句
            String text = DaoConst.UPDATE_MODEL.replace(DaoConst.TABLE_CODE, this.name);
            SqlData sql = new SqlData(DaoConst.REMOVE, text);
            sql.setSqlType(DaoConst.SQL_TYPE_REMOVE);
            sql.addCode(DaoConst.FIELD_CODE, "a." + removeField.name + "=1");
            for (TableFieldVo field : this.getFieldActionList(DaoConst.FIELD_REMOVE_UPDATE)) {
                sql.addParaConst(field.inputName, DaoConst.FIELD_CODE, ",a." + field.name + "=?");
            }
            if (versionField != null) {
                sql.addCode(DaoConst.FIELD_CODE, ",a." + versionField.name + "=1+a." + versionField.name);
            }
            addWhereField(sql, "a.", true);
            return sql;
        } else {
            String text = DaoConst.REMOVE_MODEL.replace(DaoConst.TABLE_CODE, this.name);
            SqlData sql = new SqlData(DaoConst.REMOVE, text);
            addWhereField(sql, "a.", true);
            return sql;
        }
    }

    /**
     * 生成加载的SQL语句
     *
     * @return
     */
    private SqlData releaseSqlLoad() {
        // 生成添加的SQL语句
        String text = DaoConst.LOAD_MODEL.replace(DaoConst.TABLE_CODE, this.name);
        SqlData sql = new SqlData(DaoConst.LOAD, text);
        sql.setSqlType(DaoConst.SQL_TYPE_LOAD);
        addWhereField(sql, "a.", false);
        return sql;
    }

    /**
     * 当前当前表所有字段的等于SQL语句
     *
     * @param sql
     */
    private void addWhereField(SqlData sql, String tag, boolean isRemove) {
        TableFieldVo removeField = this.getRemove();
        sql.add(this.getKey().inputName, " AND " + tag + this.getKey().name + "=?");

        // 删除时更新字段
        List<TableFieldVo> removeUpdate = this.getRemoveUpdate();
        Map<TableFieldVo, Boolean> updateField = new HashMap<>(removeUpdate.size());
        if (isRemove) {
            for (TableFieldVo field : removeUpdate) {
                updateField.put(field, true);
            }
        }

        for (TableFieldVo field : this.getFields()) {
            if (updateField.containsKey(field)) {
                continue;
            }
            sql.add(field.inputName, " AND " + tag + field.name + "=?");
        }
        if (removeField != null) {
            sql.addConst(DaoConst.REMOVE_FLAG_INPUT, " AND " + tag + removeField.name + "=0");
        }
    }

    /**
     * 生成统计的SQL语句
     *
     * @param sqlTableData 需要生成的实体
     * @param whereFields  WHERE字段
     * @param updateFields 需要增加的值的字段
     */
    public void addGroupSql(TableSqlCache sqlTableData, TableFieldString whereFields, TableFieldString updateFields) {
        sqlTableData.add(this.releaseSqlWhere(DaoConst.GROUP_QUERY, DaoConst.LOAD_MODEL, whereFields));
        sqlTableData.add(this.releaseSql(DaoConst.GROUP_ADD, DaoConst.UPDATE_MODEL,
                "{FIELD}={FIELD}+?", updateFields, new TableFieldString(this.getKey().name)));
    }

    /**
     * 生成SQL语句
     *
     * @param name
     * @param model
     * @param whereFields
     * @return
     */
    public SqlData releaseSqlWhere(String name, String model, TableFieldString whereFields) {
        return this.releaseSql(name, model, "", new TableFieldString(), whereFields);
    }

    /**
     * 生成SQL语句
     *
     * @param name        SQL语句名称
     * @param model       SQL语句模板
     * @param valueModel  值字段模板
     * @param valueFields 值字段
     * @param whereFields WHERE字段
     * @return
     */
    public SqlData releaseSql(String name, String model, String valueModel, TableFieldString valueFields, TableFieldString whereFields) {
        // 参数字段
        List<String> paraFields = new ArrayList<String>();

        // 生成添加的SQL语句
        StringBuilder sbField = new StringBuilder();
        for (String field : valueFields.getFields()) {
            if (sbField.length() > 0) {
                sbField.append(",");
            }
            sbField.append(valueModel.replace(DaoConst.FIELD_CODE, field));
            paraFields.add(field);
        }

        StringBuilder sbWhere = new StringBuilder();
        for (String field : whereFields.getFields()) {
            sbWhere.append(" AND ");
            sbWhere.append(field);
            sbWhere.append("=?");
            paraFields.add(field);
        }

        String[] fields = new String[paraFields.size()];
        fields = paraFields.toArray(fields);
        // 生成修改的SQL语句
        String sql = model.replace(DaoConst.TABLE_CODE, this.name).replace(DaoConst.FIELD_CODE, sbField.toString()).replace(DaoConst.WHERE_CODE, sbWhere.toString());
        SqlData ret = new SqlData(name, sql, fields);
        return ret;
    }

    /**
     * 生成根据某些字段不存在则保存的SQL语句
     *
     * @param tableStruct
     * @param sqlName
     * @param whereFields
     */
    public void addSaveWithSql(TableSqlCache tableStruct, String sqlName, TableFieldString whereFields) {
        tableStruct.add(this.releaseSqlWhere(sqlName, DaoConst.LOAD_MODEL, whereFields));
    }

    /**
     * 生成判断数据是否存在的SQL语句
     *
     * @param sqlTableData
     * @param sqlName
     * @param fields
     */
    public void addExist(TableSqlCache sqlTableData, String sqlName, String[] fields) {
        String text = DaoConst.LOAD_MODEL.replace(DaoConst.TABLE_CODE, this.name);
        SqlData sql = new SqlData(sqlName, text);
        sql.setSqlType(DaoConst.SQL_TYPE_EXINTS);
        sql.addPara(this.getKey().inputName, DaoConst.WHERE_CODE, " AND a." + this.getKey().name + "<>?");

        for (String fieldName : fields) {
            TableFieldVo field = this.getField(fieldName);
            sql.addParaConst(field.inputName, DaoConst.WHERE_CODE, " AND a." + field.name + "=?");
        }
        if (getRemove() != null) {
            sql.addParaConst(DaoConst.REMOVE_FLAG_INPUT, DaoConst.WHERE_CODE, " AND a." + this.getRemove().name + "=0");
        }
        sqlTableData.add(sql);
    }
}
